from __future__ import annotations

import time

from rich.console import Group
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.message import Message
from textual.widgets import Button, Static

from whosat.formatting import fmt_bytes, fmt_percent, fmt_uptime
from whosat.services.path_display import get_display_path
from whosat.theme import ThemePalette
from whosat.types import ProcessRecord


class DetailPanel(Vertical):
    class ActionRequested(Message):
        def __init__(self, sender: "DetailPanel", action: str) -> None:
            self.action = action
            super().__init__()

    class KillRequested(Message):
        def __init__(self, sender: "DetailPanel", row_key: str) -> None:
            self.row_key = row_key
            super().__init__()

    class CloseRequested(Message):
        pass

    def compose(self):
        with Horizontal(id="detail-header"):
            yield Static("", id="detail-title")
            yield Button("\u2715 close", id="close-btn")
        with VerticalScroll(id="detail-body"):
            yield Static("Select a process", id="detail-info")
        with Horizontal(id="action-row-1"):
            yield Button("PING", id="ping-btn")
            yield Button("CURL", id="curl-btn")
            yield Button("COPY IP:PORT", id="copy-ip-port-btn")
        yield Button("KILL", id="kill-btn", variant="error")

    def update_view(self, row: ProcessRecord | None, palette: ThemePalette | None = None) -> None:
        info = self.query_one("#detail-info", Static)
        title = self.query_one("#detail-title", Static)
        kill = self.query_one("#kill-btn", Button)
        ping = self.query_one("#ping-btn", Button)
        curl = self.query_one("#curl-btn", Button)
        copy_ip_port = self.query_one("#copy-ip-port-btn", Button)

        if row is None:
            title.update("")
            info.update("Select a process")
            kill.disabled = True
            ping.disabled = True
            curl.disabled = True
            copy_ip_port.disabled = True
            return

        kill.disabled = row.pid is None and not row.ports
        has_port = bool(row.ports)
        ping.disabled = not has_port
        curl.disabled = not has_port
        copy_ip_port.disabled = not has_port

        p = palette
        title.update(self._header_line(row, palette))
        info.update(self._render_detail(row, palette))

    def _render_detail(self, row: ProcessRecord, palette: ThemePalette | None):
        p = palette
        sections: list = []

        sections.append(Rule(style=p.border if p else "#1e2a38"))

        # STATUS section
        sections.append(self._section_title("STATUS", palette))
        status = row.derived_status
        dot_color = {
            "ONLINE": p.accent if p else "#00ff88",
            "WARN": p.warn if p else "#ffd060",
            "OFFLINE": p.danger if p else "#ff4466",
        }.get(status, p.text_secondary if p else "#8aa0b0")
        sections.append(self._kv_block([
            ("State", f"\u25cf {status}"),
            ("Uptime", fmt_uptime(time.time() - row.create_time) if row.create_time else "-"),
            ("Restarts", "0"),
            ("Last check", "2s ago"),
        ], palette, highlights={"State": dot_color, "Uptime": p.secondary if p else "#00d4ff"}))

        # NETWORK section
        sections.append(self._section_title("NETWORK", palette))
        bind = f"{row.ports[0].ip}:{row.ports[0].port}" if row.ports else "\u2014"
        proto = f"{row.ports[0].proto.upper()}/{row.ports[0].family.upper()}" if row.ports else "\u2014"
        sections.append(self._kv_block([
            ("Bind", bind),
            ("Local IP", row.ports[0].ip if row.ports else "-"),
            ("Protocol", proto),
            ("Connections", "\u2014"),
            ("RX / TX", "\u2014"),
        ], palette, highlights={"Bind": p.secondary if p else "#00d4ff", "Connections": p.secondary if p else "#00d4ff"}))

        # ALL OPEN PORTS section
        sections.append(self._section_title("ALL OPEN PORTS", palette))
        sections.append(self._ports_line(row, palette))

        # RESOURCES section
        sections.append(self._section_title("RESOURCES", palette))
        sections.append(self._kv_block([
            ("CPU", fmt_percent(row.cpu_percent)),
            ("Memory", f"{fmt_bytes(row.memory_bytes)} ({fmt_percent(row.memory_percent)})"),
            ("Threads", str(row.threads or "-")),
            ("FDs", str(row.fd_count or "-")),
        ], palette, highlights={"CPU": p.secondary if p else "#00d4ff", "Memory": p.secondary if p else "#00d4ff"}))

        # PATH section
        sections.append(self._section_title("PATH", palette))
        sections.append(self._wrapped_text(get_display_path(row) or "\u2014", palette))

        # CMD section
        sections.append(self._section_title("CMD", palette))
        sections.append(self._wrapped_text(row.cmdline_text or row.exe or "\u2014", palette, dim=True))

        return Group(*sections)

    def _header_line(self, row: ProcessRecord, palette: ThemePalette | None) -> Text:
        p = palette
        accent = p.accent if p else "#00ff88"
        yellow = p.warn if p else "#ffd060"
        red = p.danger if p else "#ff4466"
        text = p.text if p else "#c8d8e8"
        text2 = p.text_secondary if p else "#8aa0b0"

        source = "docker" if (row.source == "docker" or row.docker_container_id) else "sys"
        status = row.derived_status
        dot_color = {"ONLINE": accent, "WARN": yellow, "OFFLINE": red}.get(status, text2)

        t = Text(no_wrap=False)
        t.append(row.name or "process", style=f"bold {text}")
        t.append(f"  pid {row.pid if row.pid is not None else '-'}", style=text2)
        t.append("  \u25cf ", style=f"bold {dot_color}")
        t.append(status, style=f"bold {dot_color}")
        t.append(f"  {source}", style=text2)
        return t

    def _section_title(self, title: str, palette: ThemePalette | None) -> Text:
        p = palette
        text3 = p.text_dim if p else "#5a6a7a"
        border = p.border if p else "#1e2a38"
        t = Text(no_wrap=False)
        t.append(title, style=f"bold {text3}")
        t.append("\n")
        t.append("\u2500" * 30, style=border)
        return t

    def _kv_block(self, rows: list[tuple[str, str]], palette: ThemePalette | None, highlights: dict[str, str] | None = None):
        p = palette
        text3 = p.text_dim if p else "#5a6a7a"
        text2 = p.text_secondary if p else "#8aa0b0"
        highlights = highlights or {}

        tbl = Table.grid(expand=True, padding=(0, 1))
        tbl.add_column(width=12, style=text3)
        tbl.add_column(ratio=1, style=text2)
        for k, v in rows:
            val_style = highlights.get(k, text2)
            tbl.add_row(
                Text(k, style=text3),
                Text(v, style=val_style),
            )
        return tbl

    def _wrapped_text(self, text_value: str, palette: ThemePalette | None, dim: bool = False) -> Text:
        p = palette
        style = p.text_dim if p and dim else (p.text_secondary if p else "#8aa0b0")
        return Text(text_value, style=style, no_wrap=False, overflow="fold")

    def _ports_line(self, row: ProcessRecord, palette: ThemePalette | None) -> Text:
        p = palette
        cyan = p.secondary if p else "#00d4ff"
        yellow = p.warn if p else "#ffd060"
        text3 = p.text_dim if p else "#5a6a7a"
        t = Text(no_wrap=False)
        if not row.ports:
            t.append("\u2014", style=text3)
            return t
        for idx, port in enumerate(row.ports[:12]):
            if idx:
                t.append("  ")
            if port.proto == "udp":
                t.append(f" {port.port} UDP ", style=f"bold {cyan} on #0d1a22")
            else:
                t.append(f" {port.port} TCP ", style=f"bold {yellow} on #2a2413")
        return t

    def on_button_pressed(self, event: Button.Pressed) -> None:
        bid = event.button.id
        if bid == "close-btn":
            self.post_message(self.CloseRequested())
        elif bid == "kill-btn":
            self.post_message(self.KillRequested(self, "selected"))
        elif bid == "ping-btn":
            self.post_message(self.ActionRequested(self, "ping"))
        elif bid == "curl-btn":
            self.post_message(self.ActionRequested(self, "curl"))
        elif bid == "copy-ip-port-btn":
            self.post_message(self.ActionRequested(self, "copy_ip_port"))



from __future__ import annotations

import datetime as _dt

from rich.text import Text
from textual.containers import Horizontal
from textual.widgets import Static

from whosat.formatting import fmt_bytes, fmt_percent, fmt_uptime
from whosat.theme import ThemePalette
from whosat.types import AppSnapshot

# Block chars from lowest to tallest (8 levels)
_VBLOCKS = " ▁▂▃▄▅▆▇█"

_MAX_SUB_LEN = 22  # truncate sub-info strings


class SysInfoItem(Static):
    """Vertical 3-line block: label, value, sub-info."""

    def __init__(self, item_id: str, **kwargs):
        super().__init__(id=f"si-{item_id}", **kwargs)

    def set_content(
        self,
        label: str,
        value: str | Text,
        sub: str | Text,
        val_color: str = "#c8d8e8",
        text3: str = "#5a6a7a",
    ) -> None:
        t = Text(no_wrap=True)
        t.append(label.upper(), style=text3)
        t.append("\n")
        if isinstance(value, Text):
            t.append_text(value)
        else:
            t.append(value, style=f"bold {val_color}")
        t.append("\n")
        if isinstance(sub, Text):
            t.append_text(sub)
        else:
            s = sub if len(sub) <= _MAX_SUB_LEN else sub[: _MAX_SUB_LEN - 1] + "\u2026"
            t.append(s, style=text3)
        self.update(t)


class SysInfoBar(Horizontal):
    """System info bar with 7 items separated by CSS border-right dividers."""

    def compose(self):
        yield SysInfoItem("hostname")
        yield SysInfoItem("cpu")
        yield SysInfoItem("memory")
        yield SysInfoItem("uptime")
        yield SysInfoItem("containers")
        yield SysInfoItem("processes")
        yield SysInfoItem("localip", classes="si-last")

    def update_view(
        self, snapshot: AppSnapshot | None, palette: ThemePalette | None = None
    ) -> None:
        p = palette
        if snapshot is None:
            return
        s = snapshot.system
        accent = p.accent if p else "#00ff88"
        cyan = p.secondary if p else "#00d4ff"
        purple = p.docker if p else "#b084ff"
        yellow = p.warn if p else "#ffd060"
        text = p.text if p else "#c8d8e8"
        text3 = p.text_dim if p else "#5a6a7a"

        # 1. HOSTNAME
        os_short = _shorten_os(s.os_version)
        self.query_one("#si-hostname", SysInfoItem).set_content(
            "HOSTNAME", s.hostname, os_short, val_color=accent, text3=text3,
        )

        # 2. CPU (cores + load combined)
        core_count = max(1, len(s.per_core_percent))
        cpu_color = yellow if s.cpu_percent > 30 else accent
        # Value line: "20 cores · 34%"
        cpu_val = Text(no_wrap=True)
        cpu_val.append(f"{core_count} cores", style=f"bold {cyan}")
        cpu_val.append(" \u00b7 ", style=text3)
        cpu_val.append(fmt_percent(s.cpu_percent), style=f"bold {cpu_color}")
        # Sub line: per-core bars
        core_bars = _per_core_bars(s.per_core_percent, p)
        self.query_one("#si-cpu", SysInfoItem).set_content(
            "CPU", cpu_val, core_bars, text3=text3,
        )

        # 3. MEMORY
        mem_bar = _gradient_mem_bar(s.memory_used_bytes, s.memory_total_bytes, p)
        mem_val = f"{fmt_bytes(s.memory_used_bytes)} / {fmt_bytes(s.memory_total_bytes)}"
        self.query_one("#si-memory", SysInfoItem).set_content(
            "MEMORY", mem_val, mem_bar, val_color=cyan, text3=text3,
        )

        # 4. UPTIME
        boot_dt = _dt.datetime.now() - _dt.timedelta(seconds=s.uptime_seconds)
        since = f"since {boot_dt.strftime('%b %d')}"
        self.query_one("#si-uptime", SysInfoItem).set_content(
            "UPTIME", fmt_uptime(s.uptime_seconds), since, val_color=accent, text3=text3,
        )

        # 5. CONTAINERS
        self.query_one("#si-containers", SysInfoItem).set_content(
            "CONTAINERS", f"{s.docker_running} running",
            f"{s.docker_stopped} stopped", val_color=purple, text3=text3,
        )

        # 6. PROCESSES
        self.query_one("#si-processes", SysInfoItem).set_content(
            "PROCESSES", str(s.total_processes),
            f"{s.processes_with_ports} w/ ports", val_color=text, text3=text3,
        )

        # 7. LOCAL IP
        primary_ip = s.local_ips[0] if s.local_ips else "-"
        ifaces = ", ".join(s.local_ips[1:3]) if len(s.local_ips) > 1 else "lo"
        self.query_one("#si-localip", SysInfoItem).set_content(
            "LOCAL IP", primary_ip, ifaces, val_color=cyan, text3=text3,
        )


# ── helpers ──────────────────────────────────────────────────────


def _shorten_os(os_version: str) -> str:
    """Abbreviate long OS strings to fit in the SysInfo bar."""
    v = os_version
    if "microsoft" in v.lower() and "WSL" in v:
        return "Linux WSL2"
    if "microsoft" in v.lower():
        return "Linux WSL"
    parts = v.split()
    if len(parts) >= 2 and parts[0] == "Linux":
        kern = parts[1].split("-")[0]
        segs = kern.split(".")
        short_kern = ".".join(segs[:2]) if len(segs) > 2 else kern
        return f"Linux {short_kern}"
    if len(v) > _MAX_SUB_LEN:
        return v[: _MAX_SUB_LEN - 1] + "\u2026"
    return v


def _per_core_bars(per_core: list[float], p: ThemePalette | None) -> Text:
    accent = p.accent if p else "#00ff88"
    yellow = p.warn if p else "#ffd060"
    red = p.danger if p else "#ff4466"
    vals = per_core if per_core else [0.0] * 4
    t = Text(no_wrap=True)
    for v in vals[:20]:
        idx = min(8, max(0, int(v / 100.0 * 8)))
        char = _VBLOCKS[idx] if idx > 0 else "\u2581"
        if v > 70:
            color = red
        elif v > 40:
            color = yellow
        else:
            color = accent
        t.append(char, style=color)
    return t


def _gradient_mem_bar(used: int, total: int, p: ThemePalette | None) -> Text:
    cyan = p.secondary if p else "#00d4ff"
    purple = p.docker if p else "#b084ff"
    bg3 = p.panel_alt if p else "#1a2030"
    total = max(1, total)
    ratio = min(1.0, max(0.0, used / total))
    bar_len = 12
    fill = int(bar_len * ratio)
    t = Text(no_wrap=True)
    for i in range(bar_len):
        if i < fill:
            if fill > 0 and i >= fill * 0.7:
                t.append("\u2588", style=purple)
            else:
                t.append("\u2588", style=cyan)
        else:
            t.append("\u2591", style=bg3)
    pct = f" {ratio * 100:.0f}%"
    t.append(pct, style=p.text_dim if p else "#5a6a7a")
    return t



from __future__ import annotations

from textual import on
from textual.containers import Horizontal
from textual.message import Message
from textual.widgets import Button, Input, Static

from whosat.types import SortField, UIState

_SORT_OPTIONS: list[SortField] = ["port", "name", "created", "cpu", "mem"]
_MEM_SORT_OPTIONS = ["mem", "cpu", "name"]


class _GroupBtn(Static):
    """Clickable button inside a connected group."""

    class Pressed(Message):
        def __init__(self, btn_id: str) -> None:
            self.btn_id = btn_id
            super().__init__()

    def on_click(self) -> None:
        self.post_message(self.Pressed(self.id or ""))


class Toolbar(Horizontal):
    class StateChanged(Message):
        def __init__(self, sender: "Toolbar", key: str, value: str | int) -> None:
            self.key = key
            self.value = value
            super().__init__()

    class RefreshRequested(Message):
        pass

    def compose(self):
        # 1. Search box — prefix + input in one bordered container
        with Horizontal(id="search-wrap"):
            yield Static("$>", id="search-prefix")
            yield Input(placeholder="search process, port, ip...", id="search")

        # 2. Scope toggle — connected button group
        with Horizontal(id="scope-group"):
            yield _GroupBtn("ALL", id="scope-all")
            yield _GroupBtn("SYS", id="scope-sys", classes="grp-divider")
            yield _GroupBtn("DOCKER", id="scope-docker", classes="grp-divider")

        # 3. Sort label + cycle button
        yield Static("SORT", id="sort-label")
        yield Button("port \u25be", id="sort-cycle")

        # 4. ASC / DESC — connected button group
        with Horizontal(id="order-group"):
            yield _GroupBtn("ASC", id="order-asc")
            yield _GroupBtn("DESC", id="order-desc", classes="grp-divider")

        # 5. Spacer
        yield Static(id="tb-spacer")

        # 6. View toggle buttons
        yield Button("\u229e GROUP VIEW", id="view-group")
        yield Button("\u2261 FLAT VIEW", id="view-flat")

    def sync_from_state(self, state: UIState) -> None:
        view = state.main_view

        # Show/hide controls based on view
        scope_grp = self.query_one("#scope-group")
        view_grp_btn = self.query_one("#view-group", Button)
        view_flat_btn = self.query_one("#view-flat", Button)

        is_ports = (view == "ports")
        scope_grp.display = is_ports
        view_grp_btn.display = is_ports
        view_flat_btn.display = is_ports

        self.query_one("#sort-cycle", Button).label = f"{state.sort_by} \u25be"

        # Scope active state
        if is_ports:
            for scope in ("all", "sys", "docker"):
                btn = self.query_one(f"#scope-{scope}", _GroupBtn)
                btn.remove_class("scope-active-all", "scope-active-sys", "scope-active-docker")
                if state.scope == scope:
                    btn.add_class(f"scope-active-{scope}")

        # ASC / DESC active state
        asc_btn = self.query_one("#order-asc", _GroupBtn)
        desc_btn = self.query_one("#order-desc", _GroupBtn)
        asc_btn.remove_class("order-active")
        desc_btn.remove_class("order-active")
        if state.sort_order == "asc":
            asc_btn.add_class("order-active")
        else:
            desc_btn.add_class("order-active")

        # View buttons active state
        if is_ports:
            grp_btn = self.query_one("#view-group", Button)
            flat_btn = self.query_one("#view-flat", Button)
            grp_btn.remove_class("view-active")
            flat_btn.remove_class("view-active")
            if state.view_mode == "group":
                grp_btn.add_class("view-active")
            else:
                flat_btn.add_class("view-active")

    @on(Input.Changed, "#search")
    def _on_search_changed(self, event: Input.Changed) -> None:
        # The search key depends on the current view, but the app determines that
        self.post_message(self.StateChanged(self, "search_query", event.value))

    def set_main_view(self, view: str) -> None:
        """Update toolbar visibility for the given main view."""
        # This is called via sync_from_state which reads state.main_view
        pass

    def on__group_btn_pressed(self, event: _GroupBtn.Pressed) -> None:
        bid = event.btn_id
        if bid.startswith("scope-"):
            self.post_message(self.StateChanged(self, "scope", bid.split("-", 1)[1]))
            return
        if bid in ("order-asc", "order-desc"):
            nxt = "asc" if bid == "order-asc" else "desc"
            self.post_message(self.StateChanged(self, "sort_order", nxt))
            return

    @on(Button.Pressed)
    def _on_button_pressed(self, event: Button.Pressed) -> None:
        bid = event.button.id or ""
        if bid == "sort-cycle":
            current = str(event.button.label).replace(" \u25be", "").strip()
            idx = _SORT_OPTIONS.index(current) if current in _SORT_OPTIONS else 0
            nxt = _SORT_OPTIONS[(idx + 1) % len(_SORT_OPTIONS)]
            self.post_message(self.StateChanged(self, "sort_by", nxt))
            return
        if bid == "view-group":
            self.post_message(self.StateChanged(self, "view_mode", "group"))
            return
        if bid == "view-flat":
            self.post_message(self.StateChanged(self, "view_mode", "flat"))
            return



import unittest

from whosat.services.status import derive_row_status
from whosat.types import PortBinding, ProcessRecord


class StatusTests(unittest.TestCase):
    def test_online_with_listener(self):
        row = ProcessRecord(pid=1, name='python', ports=[PortBinding(8000, 'tcp', 'ipv4', '0.0.0.0')])
        self.assertEqual(derive_row_status(row), 'ONLINE')

    def test_warn_on_high_cpu(self):
        row = ProcessRecord(pid=1, name='python', cpu_percent=95.0)
        self.assertEqual(derive_row_status(row), 'WARN')

    def test_warn_on_high_memory(self):
        row = ProcessRecord(pid=1, name='python', memory_percent=90.0)
        self.assertEqual(derive_row_status(row), 'WARN')

    def test_offline_docker_synthetic(self):
        row = ProcessRecord(pid=None, name='svc', source='docker')
        self.assertEqual(derive_row_status(row), 'OFFLINE')


if __name__ == '__main__':
    unittest.main()
